from battetl import logger, Constants, Utils
import os
import copy
import json
import time
import psycopg2
import psycopg2.pool
import psycopg2.sql
import sqlalchemy
import sqlalchemy.orm
import numpy as np
import pandas as pd
from schema import Schema, Use, Optional, And, SchemaError

# progress bar
from tqdm import tqdm
# register the pandas extension
tqdm.pandas()


class Loader:
    def __init__(
            self,
            config: dict,
            env_path: str = os.path.join(os.getcwd(), '.env'),
            battdb_version: float = Constants.BATTDB_SCHEMA_VERSION,
    ):
        """
        An interface to load battery test data to a database.

        Parameters
        ----------
        config : dict
            A configuration file containing meta_data about the test data.
        env_path : str, optional
            Path to the environment file containing the database connection
            information. The default is '.env'. The environment file must
            contain the following fields:
                DB_TARGET - The target database to load data to.
                DB_USERNAME - The username to connect to the database.
                DB_PASSWORD - The password to connect to the database.
                DB_HOSTNAME - The hostname of the database.
                DB_PORT - The port of the database.
        battdb_version : float, optional
            The expected schema version of the target database. The default is
            Constants.BATTDB_SCHEMA_VERSION.
        """
        Utils.load_env(env_path)

        self.schema = Schema({
            'test_meta': {
                'test_name': str,
                'channel': Use(int),
                Optional('test_id'): Use(int),
                Optional('cell_id'): Use(int),
                Optional('start_date'): str,
                Optional('end_date'): str,
                Optional('ev_chamber'): Use(int),
                Optional('ev_chamber_slot'): Use(int),
                Optional('thermocouples'): Use(int),
                Optional('thermocouple_channels'): Use(int),
                Optional('comments'): str,
                Optional('project_id'): Use(int),
                Optional('test_capacity_mah'): Use(float),
                Optional('potentiostat_id'): Use(int),
                Optional('cycler_id'): Use(int),
            },
            'cycler': {
                'sn': str,
                Optional('cycler_id'): Use(int),
                Optional('cycler_type_id'): Use(int),
                Optional('calibration_date'): str,
                Optional('calibration_due_date'): str,
                Optional('location'): str,
                Optional('timezone_based'): str,
            },
            'cycler_meta': {
                'manufacturer': str,
                'model': str,
                Optional('cycler_type_id'): Use(int),
                Optional('manufacturer'): str,
                Optional('model'): str,
                Optional('datasheet'): str,
                Optional('num_channels'): Use(int),
                Optional('lower_current_limit_a'): Use(float),
                Optional('upper_current_limit_a'): Use(float),
                Optional('lower_voltage_limit_v'): Use(float),
                Optional('upper_voltage_limit_v'): Use(float),
            },
            'schedule_meta': {
                'schedule_name': str,
                'cycler_make': str,
                Optional('schedule_id'): Use(int),
                Optional('test_type'): And(
                    str,
                    lambda s: s in [
                        'ICT', 'Characterization', 'Baseline life cycling',
                        'MBC life cycling', 'MBC quick test', 'Pretest', 'HPPC'
                    ]
                ),
                Optional('cycler_make'): str,
                Optional('date_created'): str,
                Optional('created_by'): str,
                Optional('comments'): str,
                Optional('cv_voltage_threshold_mv'): Use(float),
                Optional('details'): str,
            },
            'cell': {
                'manufacturer_sn': str,
                Optional('cell_id'): Use(int),
                Optional('cell_type_id'): Use(int),
                Optional('batch_number'): str,
                Optional('label'): str,
                Optional('date_received'): str,
                Optional('comments'): str,
                Optional('date_manufactured'): str,
                Optional('manufacturer_sn'): str,
                Optional('dimensions'): str,
                Optional('weight_g'): Use(float),
                Optional('first_received_at_voltage_mv'): Use(float),
            },
            'cell_meta': {
                'manufacturer': str,
                'manufacturer_pn': str,
                Optional('cell_type_id'): Use(int),
                Optional('form_factor'): str,
                Optional('capacity_mah'): Use(float),
                Optional('chemistry'): str,
                Optional('dimensions'): str,
                Optional('datasheet'): str,  # File path
            },
            'customers': {
                'customer_name': str,
            },
            'projects': {
                'project_name': str,
            }
        })

        assert (self.__validate_config(config))
        assert (self.__create_connection())
        assert (self.__check_battdb_version(battdb_version))

    def load_test_data(self, df: pd.DataFrame, retry_cnt: int = 0) -> int:
        """
        Loads test data to the target database. Only loads data past the
        latest `unixtime_s` field that already exists in the `test_data` table.

        Parameters
        ----------
        df : pd.DataFrame
            Data Frame containing test data to load to database.
        retry_cnt : int, optional
            The number of times to retry loading data to the database.

        Returns
        -------
        num_rows_loaded : int
            The number of rows inserted into the `data` table.
        """
        logger.info('Loading test data to database')

        num_rows_loaded = 0

        df_copy = df.copy(deep=True)

        try:
            if retry_cnt > 0:
                logger.info(f'Retry count: {retry_cnt}')
                # Re-create connection
                self.__create_connection()

            latest_unixtime_s = self.__lookup_latest_unixtime()
            if latest_unixtime_s:
                logger.info(f'Data rows to load: {df_copy.shape[0]}')
                logger.info(
                    f'Found previous data with latest unixtime_s: {latest_unixtime_s}')
                df_copy = df_copy[df_copy['unixtime_s'] > latest_unixtime_s]
                logger.info(f'New data rows to load: {df_copy.shape[0]}')

            if df_copy.shape[0] > 0:
                # Move fields to other_details
                df_copy = self.__create_other_details(
                    df_copy, Constants.COLUMNS_TEST_DATA)

                for column in df_copy.columns:
                    if column not in Constants.COLUMNS_TEST_DATA:
                        logger.debug(
                            f'Dropping column {column} from DataFrame')
                        df_copy = df_copy.drop([column], axis=1)

                test_id = self._lookup_test_id()
                if not test_id:
                    test_id = self.__insert_test_meta()
                df_copy['test_id'] = np.ones(
                    df_copy.shape[0], dtype=np.uint16) * test_id

                num_rows_loaded += self._load_dataframe(
                    df=df_copy, target_table='test_data')

                # Update test_meta start_date and end_date
                self.__update_first_and_last_recorded_datetime(test_id)
            else:
                logger.info(
                    f'No new data to load to test_data table for test_id')

        except Exception as e:
            logger.error('Error loading test data')
            logger.error(e)
            if retry_cnt < Constants.DATABASE_MAX_RETRIES:
                logger.info(
                    f'Retrying load_test_data() {retry_cnt+1}/{Constants.DATABASE_MAX_RETRIES}')
                retry_delay = min(Constants.DATABASE_RETRY_DELAY *
                                  (retry_cnt+1), Constants.DATABASE_MAX_RETRY_DELAY)
                logger.info(f'Retry delay: {retry_delay} seconds')
                time.sleep(retry_delay)

                num_rows_loaded += self.load_test_data(
                    df=df_copy, retry_cnt=retry_cnt+1)
            else:
                logger.error(f'Exceeded max retries for load_test_data()')
                raise e

        return num_rows_loaded

    def load_cycle_stats(self, df: pd.DataFrame) -> int:
        """
        Loads cycle stats to the target database. Any cycles that already exist in the 
        database that overlap with the new cycle data will be overwritten. 

        Parameters
        ----------
        df : pd.DataFrame
            Data Frame containing test data to load to database.

        Returns
        -------
        num_rows_loaded : int
            The number of rows inserted into the `test_data` table.
        """
        logger.info('Loading cycle stats to database')

        df_copy = df.copy(deep=True)

        # Move fields to other_details
        df_copy = self.__create_other_details(
            df_copy, Constants.COLUMNS_CYCLE_STATS)

        for column in df_copy.columns:
            if column not in Constants.COLUMNS_CYCLE_STATS:
                logger.debug(f'Dropping column {column} from DataFrame')
                df_copy = df_copy.drop([column], axis=1)

        test_id = self._lookup_test_id()
        if not test_id:
            test_id = self.__insert_test_meta()

        # Add the test_id as a column to the stats dataframe.
        df_copy['test_id'] = np.ones(
            df_copy.shape[0], dtype=np.uint16) * test_id

        # Get latest cycle_stats_id for the specific test_id
        with self._conn.cursor() as cursor:
            stmt = psycopg2.sql.SQL("""
                SELECT
                    cycle_stats_id, cycle
                FROM
                    test_data_cycle_stats
                WHERE
                    test_id = {test_id}
                AND 
                    cycle >= {newest_cycle}
                ORDER BY cycle_stats_id ASC
            """).format(
                test_id=psycopg2.sql.Literal(str(test_id)),
                newest_cycle=psycopg2.sql.Literal(
                    str(int(df_copy.cycle.iloc[0])))
            )
            cursor.execute(stmt)
            result = cursor.fetchall()
        if result:
            duplicate_stats_id_bounds = [result[0][0], result[-1][0]]
            duplicate_cycle_range = [result[0][1], result[-1][1]]
            logger.info(
                f'Data through cycle {duplicate_cycle_range[-1]} already exists for test_id {test_id}')

        # Insert the new cycle stats data.
        num_rows_inserted = self._load_dataframe(
            df=df_copy, target_table='test_data_cycle_stats')

        # If new data was inserted for cycles that previously existed then delete the old data.
        if num_rows_inserted and result:
            with self._conn.cursor() as cursor:
                stmt = psycopg2.sql.SQL("""
                    DELETE FROM 
                        test_data_cycle_stats
                    WHERE 
                        test_id = {test_id}
                    AND
                        cycle_stats_id 
                    BETWEEN
                        {duplicate_lower_bound}
                    AND 
                        {duplicate_upper_bound}
                """).format(
                    test_id=psycopg2.sql.Literal(str(test_id)),
                    duplicate_lower_bound=psycopg2.sql.Literal(
                        str(duplicate_stats_id_bounds[0])),
                    duplicate_upper_bound=psycopg2.sql.Literal(
                        str(duplicate_stats_id_bounds[1])),
                )
                cursor.execute(stmt)
            logger.info(
                f'Deleted old cycle_stats data for test_id {test_id} where cycle was greater than {df_copy.cycle.iloc[0]}')

        return num_rows_inserted

    def __validate_config(self, config: dict) -> bool:
        """
        Validates the passed configuration file for Loader class. 

        Parameters
        ----------
        config : dict
            Configuration for Loader class as defined in class description

        Returns
        -------
        valid : bool
            True if config is valid, False otherwise.
        """
        if not config:
            raise ValueError('Config cannot be empty.')
        if not 'meta_data' in config:
            raise ValueError('Config must contain meta_data.')

        try:
            self.schema.validate(config['meta_data'])
        except SchemaError as e:
            raise ValueError(e)

        self.config = config['meta_data']
        return True

    def __get_db_credentials(self):
        """
        Fetches database credentials from .env file. 

        Requires that the following environment variables are set:
            DB_TARGET,
            DB_USERNAME,
            DB_PASSWORD,
            DB_HOSTNAME,
            DB_PORT
        """
        # Validate that all required environment variables are set.
        if not os.getenv('DB_TARGET'):
            raise ValueError('DB_TARGET not set in environment variables.')
        if not os.getenv('DB_USERNAME'):
            raise ValueError('DB_USERNAME not set in environment variables.')
        if not os.getenv('DB_PASSWORD'):
            raise ValueError('DB_PASSWORD not set in environment variables.')
        if not os.getenv('DB_HOSTNAME'):
            raise ValueError('DB_HOSTNAME not set in environment variables.')
        if not os.getenv('DB_PORT'):
            raise ValueError('DB_PORT not set in environment variables.')

    def __create_connection(self) -> bool:
        """
        Creates connection to the target database specified in the config. 
        `self._conn` is the psycopg2 use for queries and limited meta insertion.
        `self.engine `is an sqlalchemy Engine instance used by Pandas for data
        frame insertion.

        The database credentials are as follows:
            - DB_TARGET - The name of the database
            - DB_USERNAME - The username to login with.
            - DB_PASSWORD- The password for the username.
            - DB_HOSTNAME - The host name to connect to. 
            - DB_PORT - The port to connect to.

        The following are optional:
            - DB_SSLMODE - Defaults to 'prefer'. 'disable', 'allow', 'prefer', 'require', 'verify-ca', 'verify-full'
            - DB_SSLROOTCERT - Path to the root certificate file.
            - DB_SSLCERT - Path to the certificate file.
            - DB_SSLKEY - Path to the key file.

        Returns
        -------
        success : bool
            True or False based on whether the database connections were created.
        """
        self.__get_db_credentials()

        success = False
        try:
            postgreSQL_pool = psycopg2.pool.SimpleConnectionPool(
                minconn=1,
                maxconn=2,
                user=os.getenv('DB_USERNAME'),
                password=os.getenv('DB_PASSWORD'),
                host=os.getenv('DB_HOSTNAME'),
                port=os.getenv('DB_PORT'),
                database=os.getenv('DB_TARGET'),
                sslmode=os.getenv('DB_SSLMODE', 'prefer'),
                sslrootcert=os.getenv('DB_SSLROOTCERT', None),
                sslcert=os.getenv('DB_SSLCERT', None),
                sslkey=os.getenv('DB_SSLKEY', None),
            )
            self._conn = postgreSQL_pool.getconn()
            self._conn.autocommit = True
            self.engine = sqlalchemy.create_engine(
                'postgresql+psycopg2://', creator=postgreSQL_pool.getconn)
            success = True
            logger.info(
                f'Created connection to database {os.getenv("DB_TARGET")}')
        except psycopg2.OperationalError as e:
            logger.error(f'psycopg2 error creating connection to database {os.getenv("DB_TARGET")}!',
                         exc_info=True)
            logger.error(e)
        except sqlalchemy.exc.SQLAlchemyError as e:
            logger.error(f'sqlalchemy error creating connection to database {os.getenv("DB_TARGET")}!',
                         exc_info=True)
            logger.error(e)
        except Exception as e:
            logger.error(f'Unexpected error creating connection to database {os.getenv("DB_TARGET")}!',
                         exc_info=True)
            logger.error(e)

        return success

    def __check_battdb_version(self, battdb_version: float) -> bool:
        """
        Checks the schema version of the target database to ensure it is compatible

        Parameters
        ----------
        battdb_version : float
            The expected schema version of the target database.

        Returns
        -------
        valid : bool
            True if the schema version is valid, False otherwise.
        """
        valid = False
        with self._conn.cursor() as cursor:
            stmt = psycopg2.sql.SQL("""
                SELECT
                    version
                FROM
                    flyway_schema_history
                ORDER BY
                    version::float
                DESC
                LIMIT 1
            """)
            cursor.execute(stmt)
            result = cursor.fetchone()

        if result:
            version = float(result[0])
            if version == battdb_version:
                valid = True
                logger.info(
                    f'BattDB schema version {result[0]} found in database')
            else:
                if version > battdb_version:
                    err = f'BattDB schema version {result[0]} found in database is newer than expected version {battdb_version}. Please update BattETL.'
                    logger.error(err)
                else:
                    err = f'BattDB schema version {result[0]} found in database is older than expected version {battdb_version}. Please update BattDB.'
                    logger.error(err)
        else:
            logger.error(f'No BattDB schema version found in database')

        return valid

    def __create_other_details(self, df: pd.DataFrame, table_columns: set) -> pd.DataFrame:
        """
        Creates the other_details column in the DataFrame. This column contains
        all fields that are not in the target table.

        Parameters
        ----------
        df : pd.DataFrame
            DataFrame to create other_details column
        table_columns : set
            Set of columns in the target table

        Returns
        -------
        df : pd.DataFrame
            DataFrame with other_details column
        """
        other_details_columns = set(df.columns) - set(table_columns)

        if other_details_columns:
            logger.info(
                f'Move fields to other_details: {", ".join(other_details_columns)}')
            df['other_details'] = df.progress_apply(
                lambda row: json.dumps({
                    c: row[c] for c in other_details_columns if not pd.isnull(row[c])
                }),
                axis=1,
            )

        return df

    def _lookup_test_id(self) -> int:
        """
        Looks up the test_id in the target database based on the test_name defined
        within the config. If no entry exists then None is returned.

        Returns
        -------
        test_id : int
            The test_id from the target database. None if the test_id does not exist
        """
        with self._conn.cursor() as cursor:
            stmt = psycopg2.sql.SQL("""
                SELECT
                    test_id
                FROM
                    test_meta
                WHERE
                    test_name = {test_name}
            """).format(
                test_name=psycopg2.sql.Literal(
                    self.config['test_meta']['test_name'])
            )
            cursor.execute(stmt)
            test_id = cursor.fetchone()

        if test_id:
            test_id = test_id[0]

        logger.debug(f'Lookup test_id: {test_id}')
        return test_id

    def __lookup_cell_type_id(self) -> int:
        """
        Looks up the cell_id in the target database based on combination
        manufacturer and  manufacturer_sn as defined within the config. 
        If no entry exists then None is returned.

        Returns
        -------
        cell_id : int
            The cell_id from the target database. None if the test_id does not exist
        """

        with self._conn.cursor() as cursor:
            stmt = psycopg2.sql.SQL("""
                SELECT
                    cell_type_id
                FROM
                    cells_meta
                WHERE
                    manufacturer = {manufacturer}
                AND 
                    manufacturer_pn = {manufacturer_pn}
            """).format(
                manufacturer=psycopg2.sql.Literal(
                    self.config['cell_meta']['manufacturer']),
                manufacturer_pn=psycopg2.sql.Literal(
                    self.config['cell_meta']['manufacturer_pn'])
            )
            cursor.execute(stmt)
            cell_type_id = cursor.fetchone()

        # Pull value out of Tuple
        if cell_type_id:
            cell_type_id = cell_type_id[0]

        logger.debug(f'Lookup cell_type_id: {cell_type_id}')
        return cell_type_id

    def _lookup_cell_id(self) -> int:
        """
        Looks up the cell_id in the target database based on manufacturer,
        manufacturer_pn, and manufacturer_sn as defined within the config. 
        If no entry exists then None is returned.

        In future, consider making nested query rather than two queries to 
        speed up.

        Returns
        -------
        cell_id : int
            The cell_id from the target database. None if the test_id does not exist
        """

        cell_type_id = self.__lookup_cell_type_id()
        if not cell_type_id:
            return None

        with self._conn.cursor() as cursor:
            stmt = psycopg2.sql.SQL("""
                SELECT
                    cell_id
                FROM
                    cells
                WHERE
                    cell_type_id = {cell_type_id}
                AND 
                    label = {label}
                AND
                    manufacturer_sn = {manufacturer_sn}
            """).format(
                cell_type_id=psycopg2.sql.Literal(
                    cell_type_id),
                label=psycopg2.sql.Literal(
                    self.config['cell']['label']),
                manufacturer_sn=psycopg2.sql.Literal(
                    self.config['cell']['manufacturer_sn'])
            )
            cursor.execute(stmt)
            cell_id = cursor.fetchone()

        if cell_id:
            cell_id = cell_id[0]

        logger.debug(f'Lookup cell_id: {cell_id}')
        return cell_id

    # TODO: Modify query so that date created is used in conjunction with schedule_name.
    def __lookup_schedule_id(self) -> int:
        """
        Looks up the schedule_id in the target database based on the 
        schedule schedule_name. Returns None if the schedule_id does not exist. 

        Returns
        -------
        schedule_id : int
            The schedule_id from the target database. None if the test_id does not exist
        """

        with self._conn.cursor() as cursor:
            stmt = psycopg2.sql.SQL("""
                SELECT
                    schedule_id
                FROM
                    schedule_meta
                WHERE
                    schedule_name = {schedule_name}
            """).format(
                schedule_name=psycopg2.sql.Literal(
                    self.config['schedule_meta']['schedule_name'])
            )
            cursor.execute(stmt)
            schedule_id = cursor.fetchone()

        # Pull value out of Tuple
        if schedule_id:
            schedule_id = schedule_id[0]

        logger.debug(f'Lookup schedule_id: {schedule_id}')
        return schedule_id

    def __lookup_cycler_type_id(self) -> int:
        """
        Looks up the cycler_type id in the target database based on the combination
        of manufacturer and model defined in the config.

        Returns
        -------
        cycler_type_id : int
            The cycler_type_id from the target database. None if the cycler_type_id does not exist.
        """

        with self._conn.cursor() as cursor:
            stmt = psycopg2.sql.SQL("""
                SELECT
                    cycler_type_id
                FROM
                    cyclers_meta
                WHERE
                    manufacturer = {manufacturer}
                AND 
                    model = {model}
            """).format(
                manufacturer=psycopg2.sql.Literal(
                    self.config['cycler_meta']['manufacturer']),
                model=psycopg2.sql.Literal(
                    self.config['cycler_meta']['model'])
            )
            cursor.execute(stmt)
            cycler_type_id = cursor.fetchone()

        # Pull value out of Tuple
        if cycler_type_id:
            cycler_type_id = cycler_type_id[0]

        logger.debug(f'Lookup cycler_type_id: {cycler_type_id}')
        return cycler_type_id

    def _lookup_cycler_id(self) -> int:
        """
        Looks up the cycler_id in the target database based on manufacturer,
        model, and cycler sn as defined within the config. 
        If no entry exists then None is returned.

        Returns
        -------
        cell_id : int
            The cell_id from the target database. None if the test_id does not exist
        """

        cycler_type_id = self.__lookup_cycler_type_id()
        if not cycler_type_id:
            return None

        with self._conn.cursor() as cursor:
            stmt = psycopg2.sql.SQL("""
                SELECT
                    cycler_id
                FROM
                    cyclers
                WHERE
                    cycler_type_id = {cycler_type_id}
                AND 
                    sn = {sn}
            """).format(
                cycler_type_id=psycopg2.sql.Literal(
                    cycler_type_id),
                sn=psycopg2.sql.Literal(
                    self.config['cycler']['sn'])
            )
            cursor.execute(stmt)
            cycler_id = cursor.fetchone()

        if cycler_id:
            cycler_id = cycler_id[0]

        logger.debug(f'Lookup cycler_id: {cycler_id}')
        return cycler_id

    def __lookup_customer_id(self) -> int:
        """
        Looks up the customer_id for the customer name specified in the config.

        If no entry exists then None is returned.

        Returns
        -------
        customer_id : str
            The customer_id for the customer specified in the config.
        """
        customer_name = self.config['customers'].get('customer_name')

        if not customer_name:
            logger.warning('No customer_name specified in config')
            return None

        with self._conn.cursor() as cursor:
            stmt = psycopg2.sql.SQL("""
                SELECT
                    customer_id
                FROM
                    customers
                WHERE
                    customer_name = {customer_name}
            """).format(
                customer_name=psycopg2.sql.Literal(
                    customer_name)
            )
            cursor.execute(stmt)
            customer_id = cursor.fetchone()

        if customer_id:
            customer_id = customer_id[0]
            logger.info(f'customer_id for {customer_name} is {customer_id}')

        return customer_id

    def __lookup_project_id(self) -> int:
        """
        Looks up the project_id for the project name specified in the config.

        If no entry exists then None is returned.

        Returns
        -------
        project_id : str
            The project_id for the project specified in the config.
        """
        project_name = self.config['projects'].get('project_name')

        if not project_name:
            logger.warning('No project_name specified in config')
            return None

        with self._conn.cursor() as cursor:
            stmt = psycopg2.sql.SQL("""
                SELECT
                    project_id
                FROM
                    projects
                WHERE
                    project_name = {project_name}
            """).format(
                project_name=psycopg2.sql.Literal(
                    project_name)
            )
            cursor.execute(stmt)
            project_id = cursor.fetchone()

        if project_id:
            project_id = project_id[0]
            logger.info(f'project_id for {project_name} is {project_id}')

        return project_id

    def __insert_test_meta(self) -> int:
        """
        Inserts a new entry in test_meta table based on info in config.

        Returns
        -------
        test_id : int
            The test_id for the newly inserted test meta. None if insert failed.
        """

        cell_id = self._lookup_cell_id()
        if not cell_id:
            logger.info(
                f'No cell_id exists for {json.dumps(self.config["cell"])}, creating new entry')
            cell_id = self.__insert_cell()

        schedule_id = self.__lookup_schedule_id()
        if not schedule_id:
            logger.info(
                f'No schedule_id exists for {json.dumps(self.config["schedule_meta"])}, creating new entry')
            schedule_id = self.__insert_schedule_meta()

        cycler_id = self._lookup_cycler_id()
        if not cycler_id:
            logger.info(
                f'No cycler_id exists for {json.dumps(self.config["cycler"])}, creating new entry')
            cycler_id = self._insert_cycler()

        project_id = self.__lookup_project_id()
        if not project_id:
            logger.info(
                f'No project_id exists for {json.dumps(self.config["projects"])}, creating new entry')
            project_id = self.__insert_project()

        upload_dict = copy.deepcopy(self.config['test_meta'])
        upload_dict['schedule_id'] = schedule_id
        upload_dict['cycler_id'] = cycler_id
        upload_dict['cell_id'] = cell_id
        upload_dict['project_id'] = project_id

        logger.debug(f'Inserting test_meta: {json.dumps(upload_dict)}')
        return self._perform_insert(target_table='test_meta', dict_to_load=upload_dict, pk_id_col='test_id')

    def __insert_cell(self) -> int:
        """
        Inserts a new entry in `cells` table based on info in config.

        Returns
        -------
        cell_id : int
            The cell_id for the newly inserted cell. None if the insert failed.
        """
        cell_type_id = self.__lookup_cell_type_id()
        if not cell_type_id:
            logger.info(
                f'No cell_type_id exists for {json.dumps(self.config["cell_meta"])}, creating new entry')
            cell_type_id = self.__insert_cell_meta()

        upload_dict = copy.deepcopy(self.config['cell'])
        upload_dict['cell_type_id'] = cell_type_id

        logger.debug(f'Inserting cell: {json.dumps(upload_dict)}')
        return self._perform_insert(target_table='cells', dict_to_load=upload_dict, pk_id_col='cell_id')

    def __insert_cell_meta(self) -> int:
        """
        Inserts a new entry in `cell_meta` table based on info in config.

        Returns
        -------
        cell_type_id : int
            The cell_type_id for the newly inserted cell_meta. None if the insert failed.
        """
        upload_dict = copy.deepcopy(self.config['cell_meta'])

        logger.debug(f'Inserting cell_meta: {json.dumps(upload_dict)}')

        # Check if 'datasheet' key exists in upload_dict and the file exists
        if 'datasheet' in upload_dict:
            if os.path.exists(upload_dict['datasheet']):
                logger.info(
                    f'Found datasheet file: {upload_dict["datasheet"]}')
                # Convert pdf file to binary data
                with open(upload_dict['datasheet'], 'rb') as f:
                    datasheet_data = f.read()
                # Add binary data to upload_dict
                upload_dict['datasheet'] = datasheet_data
            else:
                logger.error(
                    f'Could not find datasheet file: {upload_dict["datasheet"]}')

        return self._perform_insert(target_table='cells_meta', dict_to_load=upload_dict, pk_id_col='cell_type_id')

    def __insert_schedule_meta(self) -> int:
        """
        Inserts a new entry in `schedule_meta` table based on info in the config.

        Returns
        -------
        schedule_id : int
            The schedule_id for the newly inserted schedule meta. None if the insert failed.
        """
        upload_dict = copy.deepcopy(self.config['schedule_meta'])

        logger.debug(f'Inserting schedule_meta: {json.dumps(upload_dict)}')
        return self._perform_insert(target_table='schedule_meta', dict_to_load=upload_dict, pk_id_col='schedule_id')

    def _insert_cycler(self) -> int:
        """
        Inserts a new entry in `cycles` table based on info in config.

        Returns
        -------
        cycler_id : int
            The cycler_id for the newly inserted cycler. None if the insert failed.
        """
        cycler_type_id = self.__lookup_cycler_type_id()
        if not cycler_type_id:
            logger.info(
                f'No cycler_type_id exists for {json.dumps(self.config["cycler_meta"])}, creating new entry')
            cycler_type_id = self.__insert_cycler_meta()

        upload_dict = copy.deepcopy(self.config['cycler'])
        upload_dict['cycler_type_id'] = cycler_type_id

        logger.debug(f'Inserting cycler: {json.dumps(upload_dict)}')
        return self._perform_insert(target_table='cyclers', dict_to_load=upload_dict, pk_id_col='cycler_id')

    def __insert_cycler_meta(self) -> int:
        """
        Inserts a new entry in `cyclers_meta` table based on info in config.

        Returns
        -------
        cycler_type_id : int
            The cycler_type_id for the new inserted row.
        """
        upload_dict = copy.deepcopy(self.config['cycler_meta'])

        logger.debug(f'Inserting cycler_meta: {json.dumps(upload_dict)}')
        return self._perform_insert(target_table='cyclers_meta', dict_to_load=upload_dict, pk_id_col='cycler_type_id')

    def __insert_project(self) -> int:
        """
        Inserts a new entry in `projects` table based on info in config. If no
        customer for the project exists, an entry is created in the `customers`
        table.

        Returns
        -------
        project_id : int
            The project_id for the newly inserted project. None if the insert failed.
        """
        customer_id = self.__lookup_customer_id()
        if not customer_id:
            logger.info(
                f'No customer_id exists for {json.dumps(self.config["customers"])}, creating new entry')
            customer_id = self.__insert_customer()

        upload_dict = copy.deepcopy(self.config['projects'])
        upload_dict['customer_id'] = customer_id

        logger.debug(f'Inserting project: {json.dumps(upload_dict)}')
        return self._perform_insert(target_table='projects', dict_to_load=upload_dict, pk_id_col='project_id')

    def __insert_customer(self) -> int:
        """
        Inserts a new entry in `customers` table based on info in config.

        Returns
        -------
        customer_id : int
            The customer_id for the newly inserted customer. None if the insert failed.
        """
        upload_dict = copy.deepcopy(self.config['customers'])

        logger.debug(f'Inserting customer: {json.dumps(upload_dict)}')
        return self._perform_insert(target_table='customers', dict_to_load=upload_dict, pk_id_col='customer_id')

    def _perform_insert(self, target_table: str, dict_to_load: dict, pk_id_col: str) -> int:
        """
        Inserts the values from the dict_to_load into the `target_table`. Assumes
        columns are keys. Drops any empty keys.

        Returns
        -------
        target_table : str
            The table to perform the insert into.
        dict_to_load : dict
            The dictionary containing the key value pairs to load into the target table.
        pk_id_col : str
            The name of the column for the primary key id.
        Returns
        -------
        pk_id : int
            The primary key id for the newly inserted row. Returns None if issue with inserting rows.
        """

        # Remove any empty entries from upload dict
        dict_to_load = {k: v for k, v in dict_to_load.items() if v is not None}
        log_dict_to_load = dict(dict_to_load)
        for k, v in log_dict_to_load.items():
            # Convert binary data to string for logging
            if isinstance(v, bytes):
                log_dict_to_load[k] = '[binary data]'
        logger.debug(
            f'Inserting into {target_table} with {log_dict_to_load}')
        pk_id = None

        try:
            insert_statement = psycopg2.sql.SQL(
                "INSERT INTO {table} ({cols}) VALUES ({vals}) RETURNING {pk_id}").format(
                table=psycopg2.sql.Identifier(target_table),
                cols=psycopg2.sql.SQL(', ').join(
                    map(psycopg2.sql.Identifier, list(dict_to_load.keys()))),
                vals=psycopg2.sql.SQL(', ').join(
                    psycopg2.sql.Placeholder() * len(dict_to_load)),
                pk_id=psycopg2.sql.Identifier(pk_id_col)
            )
            with self._conn.cursor() as cursor:
                cursor.execute(insert_statement, list(dict_to_load.values()))
                result = cursor.fetchone()
            if result:
                pk_id = result[0]
                logger.info(
                    f'Inserted new data into {target_table} with pk_id_col={pk_id_col} and pk_id={pk_id}')
            else:
                logger.error(
                    f'No result returned when inserting into {target_table} with pk_id_col={pk_id_col}')
        except psycopg2.Error as e:
            logger.error(
                f'Error inserting into {target_table} with pk_id_col={pk_id_col}')
            logger.error(e)
        except Exception as e:
            logger.error(
                f'Unexpected error inserting into {target_table} with pk_id_col={pk_id_col}')
            logger.error(e)

        return pk_id

    def __lookup_latest_unixtime(self) -> int:
        """
        Looks up the latest `unixtime_s` loaded into the `test_data` table for the
        test specified in the config. If there is no entry `unixtime_s` in the 
        data table then None is returned. 

        Returns
        -------
        latest_unixtime_s : float
            The latest unixtime_s in the `test_data` table for the test.
        """

        test_id = self._lookup_test_id()
        if not test_id:
            test_name = self.config['test_meta']['test_name']
            logger.info(
                f'No test data exists for "{test_name}", OK to upload all data.')
            return None

        with self._conn.cursor() as cursor:
            cursor.execute("""
                SELECT
                    MAX(unixtime_s)
                FROM
                    test_data
                WHERE
                    test_id = %(test_id)s
            """, {
                'test_id': str(test_id)
            })
            latest_unixtime_s = cursor.fetchone()

        if latest_unixtime_s:
            latest_unixtime_s = latest_unixtime_s[0]

        logger.info(
            f'Latest unixtime_s for test_id={test_id} is {latest_unixtime_s}')

        return latest_unixtime_s

    def __lookup_latest_cycle(self) -> int:
        """
        Looks up the latest `cycle` loaded into the `test_data_cycle_stats` table
        for the test specified in the config. If there is no entry for `cycle` in the 
        table then None is returned. 

        Returns
        -------
        latest_cycle : int
            The latest cycle uploaded in the database for the test.
        """

        test_id = self._lookup_test_id()
        if not test_id:
            logger.info(
                f'No test data exists for {test_id}, OK to upload all data.')
            return None

        with self._conn.cursor() as cursor:
            cursor.execute("""
                SELECT
                    MAX(cycle)
                FROM
                    test_data_cycle_stats
                WHERE
                    test_id = %(test_id)s
            """, {
                'test_id': str(test_id)
            })
            latest_cycle = cursor.fetchone()

        if latest_cycle:
            latest_cycle = latest_cycle[0]

        logger.info(f'Latest cycle for test_id={test_id} is {latest_cycle}')

        return latest_cycle

    def _load_dataframe(self, df: pd.DataFrame, target_table: str) -> int:
        """
        Loads the passed data frame to the passed target_table in the database
        specified in the config.

        Parameters
        ----------
        df : pd.DataFrame
            Data frame to load to database.
        target_table : str
            Table to load the data frame to.

        Returns
        -------
        num_rows_inserted : int
            The number of rows inserted into the target_table.
        """
        __upload_chunk_size = 10000
        num_rows_inserted = 0

        def chunker(seq, size):
            # from http://stackoverflow.com/a/434328
            return (seq[pos:pos + size] for pos in range(0, len(seq), size))

        try:
            logger.info(f'Inserting {len(df)} rows into {target_table} table.')
            with tqdm(total=len(df)) as pbar:
                for chunk in chunker(df, __upload_chunk_size):
                    num_rows_inserted += chunk.to_sql(
                        name=target_table,
                        con=self.engine,
                        schema='public',
                        if_exists='append',
                        index=False,  # Do not include the pd table index as a column
                        # Pass multiple values in a single INSERT clause.
                        method='multi'
                    )
                    pbar.update(len(chunk))

            logger.info(
                f'Inserted {num_rows_inserted} rows into {target_table} table.')
        except psycopg2.OperationalError as e:
            logger.error(
                f'Error inserting into {target_table} table')
            logger.error(e)
        except Exception as e:
            logger.error(
                f'Unexpected error inserting into {target_table} table.')
            logger.error(e)
            if target_table == 'test_data':
                raise e

        return num_rows_inserted

    def __update_first_and_last_recorded_datetime(self, test_id):
        """
        Updates the first_recorded_datetime and last_recorded_datetime for the test

        Parameters
        ----------
        test_id : int
            The test_id for the test to update.
        """
        logger.info(
            f'Updating first_recorded_datetime and last_recorded_datetime for test_id={test_id}')
        with self._conn.cursor() as cursor:
            # "GROUP BY" is required to avoid error
            # https://stackoverflow.com/a/19602031
            cursor.execute("""
                UPDATE test_meta
                SET first_recorded_datetime = (
                    SELECT recorded_datetime
                    FROM test_data
                    WHERE test_id = %(test_id)s
                    GROUP BY test_data_id, recorded_datetime
                    ORDER BY test_data_id ASC
                    LIMIT 1
                ),
                last_recorded_datetime = (
                    SELECT recorded_datetime
                    FROM test_data
                    WHERE test_id = %(test_id)s
                    GROUP BY test_data_id, recorded_datetime
                    ORDER BY test_data_id DESC
                    LIMIT 1
                )
                WHERE test_id = %(test_id)s
            """, {
                'test_id': str(test_id)
            })
            self._conn.commit()



import pandas as pd
import json
import os
import copy
from .Loader import Loader
from battetl import logger, Constants


class QuickLoader(Loader):

    def __init__(self, file_path):

        # The config needed by the BattETL parent class
        self._file_name = os.path.basename(file_path)
        config = {
            "meta_data": {
                "test_meta": {
                    "test_name": copy.deepcopy(self._file_name),
                    "channel": "0"
                },
                "cell": {
                    "manufacturer_sn": "NA"
                },
                "cell_meta": {
                    "manufacturer": "NA",
                    "manufacturer_pn": "NA",
                },
                "schedule_meta": {
                    "schedule_name": "NA",
                    "cycler_make": "NA"
                },
                "cycler": {
                    "sn": "NA",
                },
                "cycler_meta": {
                    "manufacturer": "NA",
                    "model": "NA",
                },
                "customers": {
                    "customer_name": "NA",
                },
                "projects": {
                    "project_name": "NA",
                }
            }
        }
        super().__init__(config=config, battdb_version=Constants.BATTDB_QUICK_SCHEMA_VERSION)

    def load_test_data(self, df: pd.DataFrame, retry_cnt: int = 0) -> int:
        """
        Loads test data to the target database. Only loads data past the
        latest `unixtime_s` field that already exists in the `test_data` table.

        Parameters
        ----------
        df : pd.DataFrame
            Data Frame containing test data to load to database.
        retry_cnt : int, optional
            The number of times to retry loading data to the database.

        Returns
        -------
        num_rows_loaded : int
            The number of rows inserted into the `data` table.
        """
        # If a test_meta entry does not already exist, then create it.
        if not super()._lookup_test_id():
            assert (self._insert_quick_test_meta(file_name=self._file_name))
        return super().load_test_data(df, retry_cnt)

    def load_cycle_stats(self, df: pd.DataFrame) -> int:
        """
        Loads cycle stats to the target database. Any cycles that already exist in the 
        database that overlap with the new cycle data will be overwritten. 

        Parameters
        ----------
        df : pd.DataFrame
            Data Frame containing test data to load to database.

        Returns
        -------
        num_rows_loaded : int
            The number of rows inserted into the `test_data` table.
        """
        if not super()._lookup_test_id():
            assert (self._insert_quick_test_meta(file_name=self._file_name))
        return super().load_cycle_stats(df)

    def _insert_quick_test_meta(self, file_name):
        '''
        Creates a test_meta entry for the passed file name. The only entry populated is the test_name column, which is set to the file name.
        '''
        upload_dict = {'test_name': file_name}
        logger.debug(f'Inserting test_meta: {json.dumps(upload_dict)}')
        return self._perform_insert(target_table='test_meta', dict_to_load=upload_dict, pk_id_col='test_id')



"""
BattETL logger module
"""

import os
import logging
import datetime
from dotenv import load_dotenv
from logging.handlers import RotatingFileHandler

load_dotenv()

LOG_MAX_SIZE = 10 * 1024 * 1024  # 10 MB
STREAM_LOG_MAX_LENGTH = 1000
MODULE_DIR = os.path.abspath(os.path.dirname(__file__))
LOG_DIR = os.path.join(MODULE_DIR, "logs")
if not os.path.isdir(LOG_DIR):
    os.mkdir(LOG_DIR)
LOG_FILE_NAME = 'battetl.log'
LOG_FILE_PATH = os.path.join(LOG_DIR, LOG_FILE_NAME)

# Log format
# ----------
# YYYY-MM-DD HH:MM:SS LEVEL [MODULE:LINENO] MESSAGE
# asctime: Human-readable time when the LogRecord was created.
# levelname: Text logging level for the message ('DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL').
# module: The module (name portion of filename).
# lineno: Source line number where the logging call was issued (if available).
# message: The logged message.
logger = logging.getLogger('battetl')

log_format_stream = f'%(asctime)s [%(levelname)s][%(filename)s:%(lineno)d][%(name)s] %(message)s'
formatter_stream = logging.Formatter(fmt=log_format_stream)
handler_stream = logging.StreamHandler()
handler_stream.setFormatter(formatter_stream)
logger.addHandler(handler_stream)

log_format_file = '{ "time": "%(asctime)s", "timestamp": "%(created)f", "level": "%(levelname)s", "file": "%(filename)s:%(lineno)d", "name": "%(name)s", "thread": "%(threadName)s", "process": "%(process)d", "message": "%(message)s" }'
formatter_file = logging.Formatter(fmt=log_format_file)
handler_file = RotatingFileHandler(
    LOG_FILE_PATH, maxBytes=LOG_MAX_SIZE, backupCount=10, delay=True)
handler_file.setFormatter(formatter_file)
logger.addHandler(handler_file)


# Set log level to DEBUG if ENV is dev
env = os.getenv('ENV')
if env == 'dev':
    logger.setLevel(logging.DEBUG)
else:
    logger.setLevel(logging.INFO)



import pandas as pd
import random
import string
from psycopg2 import sql
from copy import deepcopy

from .Loader import Loader


class BattDbTestHelper(Loader):

    cell_type_id = None
    cell_id = None
    cycler_type_id = None
    cycle_id = None
    schedule_id = None
    test_id = None
    profile_id = None
    customer_id = None

    def __init__(self, config: dict):
        '''
        Various methods to help with tests involving BattDB
        '''
        super().__init__(config)
        self.config = config['meta_data']
        assert (self._Loader__create_connection())

    def create_test_db_entries(self):
        '''
        Creates various entries into the database that the tests assume exist.
        Assert to make sure the inserts actually happened. If they fail, then
        likely there is data in the database that needs to be cleared. Easiest course
        of action in that case is rebuilding the database from scratch.
        '''
        self.cell_type_id = self._Loader__insert_cell_meta()
        assert (self.cell_type_id)

        upload_dict = {**self.config['cell'],
                       **{'cell_type_id': self.cell_type_id}}
        self.cell_id = self._perform_insert(
            target_table='cells', dict_to_load=upload_dict, pk_id_col='cell_id')
        assert (self.cell_id)

        self.cycler_type_id = self._perform_insert(
            target_table='cyclers_meta', dict_to_load=self.config['cycler_meta'], pk_id_col='cycler_type_id')
        assert (self.cycler_type_id)

        upload_dict = {**self.config['cycler'],
                       **{'cycler_type_id': self.cycler_type_id}}
        self.cycler_id = self._perform_insert(
            target_table='cyclers', dict_to_load=upload_dict, pk_id_col='cycler_id')
        assert (self.cycler_id)

        self.schedule_id = self._perform_insert(
            target_table='schedule_meta', dict_to_load=self.config['schedule_meta'], pk_id_col='schedule_id')
        assert (self.schedule_id)

        upload_dict = {**self.config['customers']}
        self.customer_id = self._perform_insert(
            target_table='customers', dict_to_load=upload_dict, pk_id_col='customer_id')
        assert (self.customer_id)

        upload_dict = {**self.config['projects'],
                       'customer_id': self.customer_id}
        self.project_id = self._perform_insert(
            target_table='projects', dict_to_load=upload_dict, pk_id_col='project_id')
        assert (self.project_id)

        upload_dict = {**self.config['test_meta'],
                       **{'cycler_id': self.cycler_id,
                          'schedule_id': self.schedule_id,
                          'cell_id': self.cell_id}}
        self.test_id = self._perform_insert(
            target_table='test_meta', dict_to_load=upload_dict, pk_id_col='test_id')
        assert (self.test_id)

        upload_dict = {'test_id': self.test_id,
                       'user_id': 'test_helper',
                       'cell_type_id': self.cell_type_id}
        self.sil_id = self._perform_insert(
            target_table='sil_meta', dict_to_load=upload_dict, pk_id_col='sil_id')

        upload_dict = {'test_id': self.test_id,
                       'user_id': 'test_helper',
                       'cell_type_id': self.cell_type_id}
        self.sim_id = self._perform_insert(
            target_table='sim_meta', dict_to_load=upload_dict, pk_id_col='sim_id')
        assert (self.test_id)

    def read_first_row(self, target_table: str, pk_col_name: str) -> tuple:
        """
        Returns the first row in the target_table, sorting off of pk_col_name.

        Parameters
        ----------
        target_table : str
            The table pull the first row from
        pk_col_name : str
            The name of the column for the primary key id.

        Returns
        -------
        pk_id : int
            The primary key id for the newly inserted row. Returns None if issue with inserting rows.
        """
        with self._conn.cursor() as cursor:
            stmt = sql.SQL("""
            SELECT
                *
            FROM
                {table}
            ORDER BY
                {pk_col_name}
            ASC LIMIT 1;""").format(
                table=sql.Identifier(target_table),
                pk_col_name=sql.Identifier(pk_col_name),
            )
            cursor.execute(stmt)
            first_row = cursor.fetchone()

        return first_row

    def read_last_row(self, target_table: str, pk_col_name: str) -> tuple:
        """
        Returns the last row in the target_table, sorting off of pk_col_name.

        Parameters
        ----------
        target_table : str
            The table pull the last row from
        pk_col_name : str
            The name of the column for the primary key id.

        Returns
        -------
        pk_id : int
            The primary key id for the newly inserted row. Returns None if issue with inserting rows.
        """
        with self._conn.cursor() as cursor:
            stmt = sql.SQL("""
            SELECT
                *
            FROM
                {table}
            ORDER BY
                {pk_col_name}
            DESC LIMIT 1;""").format(
                table=sql.Identifier(target_table),
                pk_col_name=sql.Identifier(pk_col_name),
            )
            cursor.execute(stmt)
            last_row = cursor.fetchone()

        return last_row

    def load_sil_data(self, df: pd.DataFrame) -> int:
        """
        Loads the passed sil_data to the database.

        Parameters
        ----------
        df : pd.DataFrame
            Dataframe with data to load to `sil_data` table

        Returns
        -------
        num_rows_loaded : int
            The number of rows loaded to the `sil_data` table
        """
        df['sil_id'] = self.sil_id
        return self._load_dataframe(df, 'sil_data')

    def load_sim_data(self, df: pd.DataFrame) -> int:
        """
        Loads the passed sim_data to the database.

        Parameters
        ----------
        df : pd.DataFrame
            Dataframe with data to load to `sim_data` table

        Returns
        -------
        num_rows_loaded : int
            The number of rows loaded to the `sim_data` table
        """
        df['sim_id'] = self.sim_id
        return self._load_dataframe(df, 'sim_data')

    def load_df_to_db(self, df: pd.DataFrame, target_table: str) -> int:
        """
        Loads the passed data frame to the passed target_table in the database
        specified in the config.

        Parameters
        ----------
        df : pd.DataFrame
            Data frame to load to database.
        target_table : str
            Table to load the data frame to.

        Returns
        -------
        num_rows_inserted : int
            The number of rows inserted into the target_table.
        """
        return self._load_dataframe(df, target_table)

    def delete_test_db_entries(self):
        '''
        Deletes the test_db entries created in `self.create_test_db_entries`
        '''
        self.delete_entry(
            'sil_meta', 'sil_id', self.sil_id)
        self.delete_entry(
            'sim_meta', 'sim_id', self.sim_id)
        self.delete_entry(
            'test_meta', 'test_id', self.test_id)
        self.delete_entry(
            'cells', 'cell_id', self.cell_id)
        self.delete_entry(
            'cells_meta', 'cell_type_id', self.cell_type_id)
        self.delete_entry(
            'cyclers', 'cycler_id', self.cycler_id)
        self.delete_entry(
            'cyclers_meta', 'cycler_type_id', self.cycler_type_id)
        self.delete_entry(
            'schedule_meta', 'schedule_id', self.schedule_id)
        self.delete_entry(
            'projects', 'project_id', self.project_id)
        self.delete_entry(
            'customers', 'customer_id', self.customer_id)

    def delete_entry(self, target_table, pk_col_name, pk_id):
        '''
        Drops an entry from a meta table.

        Parameters
        ----------
        target_table : str
            The name of the meta table. E.g. "cells_meta"
        pk_col_name : str
            The name of the primary key column. E.g. "cell_type_id"
        pk : int
            The primary key id number. E.g 1
        '''
        with self._conn.cursor() as cursor:
            stmt = sql.SQL("""
                DELETE FROM
                    {table_name}
                WHERE
                    {pk_col_name} = {pk_id}
            """).format(
                table_name=sql.Identifier(target_table),
                pk_col_name=sql.Identifier(pk_col_name),
                pk_id=sql.Literal(pk_id)
            )
            cursor.execute(stmt)

    def delete_test_data(self):
        '''
        Deletes all data from from data tables
        '''
        data_to_delete_info = [
            ('test_data', 'test_id', self.test_id),
            ('test_data_cycle_stats', 'test_id', self.test_id),
            ('sil_data', 'sil_id', self.sil_id),
            ('sim_data', 'sim_id', self.sim_id),
        ]
        for data_info in data_to_delete_info:
            self.delete_entry(data_info[0], data_info[1], data_info[2])

    def generate_random_string(self, len=20) -> str:
        '''
        Generates a random string of length `len`.

        Parameters
        ----------
        len : int
            The desired length of the random string

        Returns
        -------
        random_string : str
            A random string of length `len`.
        '''
        return (''.join(random.choice(string.ascii_letters) for i in range(len)))



import os
import json
import pytest
import pandas as pd

from battetl import BattETL
from battetl.load import BattDbTestHelper

CONFIG_DIR = os.path.join(os.path.dirname(__file__), 'configs')


class Values:
    """
    Class to store values that are used in multiple tests

    Attributes
    ----------
    cell : BattETL
        The BattETL object that is used in the tests
    CONFIG_1 : dict
        The configuration dictionary for the test
    TEST_HELPER : BattDbTestHelper
        The BattDbTestHelper object that is used in the tests
    """
    cell: BattETL = None
    CONFIG_1: dict = None
    TEST_HELPER: BattDbTestHelper = None


@pytest.fixture(scope='module', autouse=True)
def add_database_test_entries():
    """
    Fixture to add test entries to the database before the tests are run
    and delete them after the tests are run
    """
    with open(os.path.join(CONFIG_DIR, 'config_1.json')) as config_file:
        Values.CONFIG_1 = json.load(config_file)

    Values.TEST_HELPER = BattDbTestHelper(Values.CONFIG_1)

    # Will be executed before the first test
    Values.TEST_HELPER.create_test_db_entries()
    yield Values.TEST_HELPER
    # Will be executed after the last test
    Values.TEST_HELPER.delete_test_db_entries()


@pytest.mark.battetl
class TestBattetl:
    def test_config(self, cell_config):
        config_path = os.path.join(
            os.path.dirname(__file__),
            'configs',
            'config_1.json'
        )

        Values.cell = BattETL(config_path)
        assert Values.cell.config == cell_config

    def test_extract(self):
        Values.cell.extract()
        assert Values.cell.schedule['file_name'] == 'tests/data/maccor_cycler_data/simple_data/BG_Maccor_Schedule.000'

    def test_transform(self):
        Values.cell.transform()
        assert not pd.DataFrame().equals(Values.cell.test_data)
        assert not pd.DataFrame().equals(Values.cell.cycle_stats)

    def test_transformed_test_data(self, data_column, data_row):
        # Maccor test data columns
        assert all(
            [col in Values.cell.test_data.columns for col in data_column.maccor.test_data]
        )

        # First row of transformed data
        row = Values.cell.test_data.iloc[0]
        assert row.equals(data_row.maccor.test_data_harmonized)

    def test_transformed_cycle_stats(self, data_column, data_row):
        # Maccor cycle stats columns
        assert all(
            [col in Values.cell.cycle_stats.columns for col in data_column.maccor.cycle_stats]
        )

        # First row of transformed data
        row = Values.cell.cycle_stats.iloc[0]
        assert row.equals(data_row.maccor.cycle_stats_calced)

    def test_load(self):
        Values.cell.load()

        db_last_row_test_meta = Values.TEST_HELPER.read_last_row(
            'test_meta', 'test_id')
        db_first_row_test_data = Values.TEST_HELPER.read_first_row(
            'test_data', 'test_data_id')
        db_last_row_test_data = Values.TEST_HELPER.read_last_row(
            'test_data', 'test_data_id')
        db_last_row_cycle_stats = Values.TEST_HELPER.read_last_row(
            'test_data_cycle_stats', 'cycle_stats_id')
        db_last_row_cells_meta = Values.TEST_HELPER.read_last_row(
            'cells_meta', 'cell_type_id')

        # Check that the first_recorded_datetime is the same as the first row of the test_data table
        assert db_last_row_test_meta[16] == db_first_row_test_data[8], \
            'first_recorded_datetime should equal the first row of the test_data table'
        # Check that the last_recorded_datetime is the same as the last row of the test_data table
        assert db_last_row_test_meta[17] == db_last_row_test_data[8], \
            'first_recorded_datetime should equal the first row of the test_data table'

        # Check that the voltage_mv is the same as the last row of the test_data table
        assert float(db_last_row_test_data[7]) == 4186.39, \
            'voltage_mv should equal 4186.39'

        # Check that the maccor_energy_mwh is the same as the last row of the test_data table
        assert 'maccor_energy_mwh' in db_last_row_test_data[-1], \
            'test_data other_details should have maccor_energy_mwh'

        # Check that the reported_charge_capacity_mah is the same as the last row of the cycle_stats table
        assert float(db_last_row_cycle_stats[4]) == 2743.710761367, \
            'reported_charge_capacity_mah should equal 2743.710761367'

        # Check that the maccor_charge_thermocouple_max_c is the same as the last row of the cycle_stats table
        assert 'maccor_charge_thermocouple_max_c' in db_last_row_cycle_stats[-1], \
            'cycle_stats other_details should have maccor_charge_thermocouple_max_c'

        # Check that the datasheet is load correctly
        with open(Values.cell.config['meta_data']['cell_meta']['datasheet'], 'rb') as f:
            datasheet_data = f.read()
        assert datasheet_data == db_last_row_cells_meta[7].tobytes(), \
            'datasheet should be load correctly'

        Values.TEST_HELPER.delete_test_data()



import os
import pytest
from os.path import join

from battetl.extract import Extractor
from battetl.transform import Transformer

BASE_DATA_PATH = os.path.join(os.path.dirname(__file__), 'data')
MACCOR_PATH = os.path.join(BASE_DATA_PATH, 'maccor_cycler_data')
MACCOR_SIMPLE_PATH = os.path.join(MACCOR_PATH, 'simple_data')
MACCOR_TYPE2_PATH = os.path.join(MACCOR_PATH, 'type2_data')
ARBIN_PATH = os.path.join(BASE_DATA_PATH, 'arbin_cycler_data')
ARBIN_SINGLE_PATH = os.path.join(ARBIN_PATH, 'single_data_file')
ARBIN_MULTIPLE_PATH = os.path.join(ARBIN_PATH, 'multiple_data_files')
ICT_PATH = os.path.join(ARBIN_PATH, 'ict_data_files')
CCCV_PATH = os.path.join(ARBIN_PATH, 'cccv_data_files')
RESET_CAP_PATH = os.path.join(MACCOR_PATH, 'reset_cap_data')
STEP_ORDER_PATH = os.path.join(ARBIN_PATH, 'step_order_data_files')


@pytest.mark.transform
@pytest.mark.maccor
def test_transform_maccor_test_data(data_column, data_row):
    path = join(MACCOR_SIMPLE_PATH, 'BG_Maccor_TestData - 079.pkl')
    extractor = Extractor()
    df_extracted = extractor.from_pickle(path)

    transformer = Transformer()
    transformer.transform_test_data(df_extracted)
    df_transformed = transformer.test_data

    # Maccor test data columns
    assert all(
        [col in df_transformed.columns for col in data_column.maccor.test_data]
    )

    # First row of transformed data
    row = df_transformed.iloc[0]
    assert row.equals(data_row.maccor.test_data)


@pytest.mark.transform
@pytest.mark.maccor
def test_transform_maccor_test_data_type2(data_column, data_row):
    path = join(MACCOR_TYPE2_PATH, 'BG_Maccor_Type2 - 075.pkl')
    extractor = Extractor()
    df_extracted = extractor.from_pickle(path)

    transformer = Transformer()
    transformer.transform_test_data(df_extracted)
    df_transformed = transformer.test_data

    # Maccor test data columns
    assert all(
        [col in df_transformed.columns for col in data_column.maccor.test_data_type2]
    )

    # First row of transformed data
    row = df_transformed.iloc[0]
    assert row.equals(data_row.maccor.test_data_type2)


@pytest.mark.transform
@pytest.mark.maccor
def test_transform_maccor_cycle_stats(data_column, data_row):
    path = join(MACCOR_SIMPLE_PATH, 'BG_Maccor_TestData - 079 [STATS].pkl')
    extractor = Extractor()
    df_extracted = extractor.from_pickle(path)

    transformer = Transformer()
    transformer.transform_cycle_stats(df_extracted)
    df_transformed = transformer.cycle_stats

    # Maccor cycle stats columns
    assert all(
        [col in df_transformed.columns for col in data_column.maccor.cycle_stats]
    )

    # First row of transformed data
    row = df_transformed.iloc[0]
    assert row.equals(data_row.maccor.cycle_stats)


@pytest.mark.transform
@pytest.mark.arbin
def test_transform_arbin_test_data(data_column, data_row):
    path = join(ARBIN_SINGLE_PATH,
                'BG_Arbin_TestData_Single_File_Channel_26_Wb_1.pkl')
    extractor = Extractor()
    df_extracted = extractor.from_pickle(path)

    transformer = Transformer()
    transformer.transform_test_data(df_extracted)
    df_transformed = transformer.test_data

    # Maccor test data columns
    assert all(
        [col in df_transformed.columns for col in data_column.arbin.test_data]
    )

    # First row of transformed data
    row = df_transformed.iloc[0]
    assert row.equals(data_row.arbin.test_data)


@pytest.mark.transform
@pytest.mark.arbin
def test_transform_arbin_cycle_stats(data_column, data_row):
    path = join(ARBIN_SINGLE_PATH,
                'BG_Arbin_TestData_Single_File_Channel_26_StatisticByCycle.pkl')
    extractor = Extractor()
    df_extracted = extractor.from_pickle(path)

    transformer = Transformer()
    transformer.transform_cycle_stats(df_extracted)
    df_transformed = transformer.cycle_stats

    # Maccor cycle stats columns
    assert all(
        [col in df_transformed.columns for col in data_column.arbin.cycle_stats]
    )

    # First row of transformed data
    row = df_transformed.iloc[0]
    assert row.equals(data_row.arbin.cycle_stats)


@pytest.mark.transform
@pytest.mark.maccor
def test_user_defined_test_data_transformation(data_column, data_row):
    path = join(MACCOR_SIMPLE_PATH, 'BG_Maccor_TestData - 079.pkl')
    extractor = Extractor()
    df_extracted = extractor.from_pickle(path)

    # User defined test data transformation
    def user_transform_test_data(df):
        df['voltage_v'] = df['voltage_mv'] * 1000
        return df

    transformer = Transformer(
        user_transform_test_data=user_transform_test_data)
    transformer.transform_test_data(df_extracted)
    df_transformed = transformer.test_data

    # Maccor test data columns
    assert all(
        [col in df_transformed.columns for col in data_column.maccor.test_data + ['voltage_v']]
    )

    # First row of transformed data
    row = df_transformed.iloc[0]
    maccor_test_data = data_row.maccor.test_data
    maccor_test_data['voltage_v'] = maccor_test_data['voltage_mv'] * 1000
    assert row.equals(maccor_test_data)


@pytest.mark.transform
@pytest.mark.maccor
def test_user_defined_cycle_stats_transformation(data_column, data_row):
    path = join(MACCOR_SIMPLE_PATH, 'BG_Maccor_TestData - 079 [STATS].pkl')
    extractor = Extractor()
    df_extracted = extractor.from_pickle(path)

    # User defined cycle stats transformation
    def user_transform_cycle_stats(df):
        df['test_time_m'] = df['test_time_s'] / 60
        return df

    transformer = Transformer(
        user_transform_cycle_stats=user_transform_cycle_stats)
    transformer.transform_cycle_stats(df_extracted)
    df_transformed = transformer.cycle_stats

    # Maccor cycle stats columns
    assert all(
        [col in df_transformed.columns for col in data_column.maccor.cycle_stats + ['test_time_m']]
    )

    # First row of transformed data
    row = df_transformed.iloc[0]
    maccor_cycle_stats = data_row.maccor.cycle_stats
    maccor_cycle_stats['test_time_m'] = maccor_cycle_stats['test_time_s'] / 60
    assert row.equals(maccor_cycle_stats)


@pytest.mark.transform
@pytest.mark.maccor
@pytest.mark.stats
def test_calc_maccor_cycle_stats():
    data_path = join(MACCOR_SIMPLE_PATH, 'BG_Maccor_TestData - 079.pkl')
    stat_path = join(MACCOR_SIMPLE_PATH,
                     'BG_Maccor_TestData - 079 [STATS].pkl')
    schedule_path = join(MACCOR_SIMPLE_PATH, 'BG_Maccor_Schedule.000')

    extractor = Extractor()
    df_data_raw = extractor.from_pickle(data_path)
    df_stats_raw = extractor.from_pickle(stat_path)
    schedule = extractor.schedule_from_files([schedule_path])

    transformer = Transformer()
    transformer.transform_test_data(df_data_raw)
    transformer.transform_cycle_stats(df_stats_raw)

    transformer.calc_cycle_stats(
        schedule['steps'], cv_voltage_threshold_mv=4195)

    # Make sure reported/calculated charge and discharge capacity are within 1% of each other.
    assert ((0.01 > abs(transformer.cycle_stats.reported_discharge_capacity_mah /
                        transformer.cycle_stats.calculated_discharge_capacity_mah - 1).iloc[1:-1])).all()

    assert ((0.01 > abs(transformer.cycle_stats.reported_charge_capacity_mah /
                        transformer.cycle_stats.calculated_charge_capacity_mah - 1).iloc[1:-1])).all()


@pytest.mark.transform
@pytest.mark.arbin
@pytest.mark.stats
def test_calc_arbin_cycle_stats():
    data_path = join(ARBIN_SINGLE_PATH,
                     'BG_Arbin_TestData_Single_File_Channel_26_Wb_1.pkl')
    stat_path = join(
        ARBIN_SINGLE_PATH, 'BG_Arbin_TestData_Single_File_Channel_26_StatisticByCycle.pkl')
    schedule_path = join(
        ARBIN_SINGLE_PATH, 'BG_25R_Characterization+BG_25R.sdx')

    extractor = Extractor()
    df_data_raw = extractor.from_pickle(data_path)
    df_stats_raw = extractor.from_pickle(stat_path)
    schedule = extractor.schedule_from_files([schedule_path])

    transformer = Transformer()
    transformer.transform_test_data(df_data_raw)
    transformer.transform_cycle_stats(df_stats_raw)

    transformer.calc_cycle_stats(
        schedule['steps'], cv_voltage_threshold_mv=4195)

    # Make sure reported/calculated charge and discharge capacity are within 1% of each other.
    assert ((0.01 > abs(transformer.cycle_stats.reported_discharge_capacity_mah /
                        transformer.cycle_stats.calculated_discharge_capacity_mah - 1).iloc[1:-1])).all()

    assert ((0.01 > abs(transformer.cycle_stats.reported_charge_capacity_mah /
                        transformer.cycle_stats.calculated_charge_capacity_mah - 1).iloc[1:-1])).all()


@pytest.mark.transform
@pytest.mark.arbin
@pytest.mark.stats
def test_ict_data():
    data_path = join(
        ICT_PATH, 'BG_Arbin_TestData_ICT_Cell3_Channel_4_Wb_1.CSV')
    stat_path = join(
        ICT_PATH, 'BG_Arbin_TestData_ICT_Cell3_Channel_4_StatisticByCycle.CSV')
    schedule_path = join(
        ICT_PATH, 'BG_Arbin_TestData_ICT+BG_Arbin_TestData_xxxxx.sdx')

    extractor = Extractor()
    extractor.data_from_files([data_path])
    extractor.data_from_files([stat_path])
    schedule = extractor.schedule_from_files([schedule_path])

    transformer = Transformer()
    transformer.transform_test_data(extractor.raw_test_data)
    transformer.transform_cycle_stats(extractor.raw_cycle_stats)

    transformer.calc_cycle_stats(
        schedule['steps'], cv_voltage_threshold_mv=4195)

    # Make sure reported/calculated stats are within 1% of each other.
    # TODO: Add additional criteria for ICT data (charge_time, discharge_time, etc.)
    assert ((0.01 > abs(transformer.cycle_stats.reported_discharge_capacity_mah /
                        transformer.cycle_stats.calculated_discharge_capacity_mah - 1).iloc[1:-1])).all()
    assert ((0.01 > abs(transformer.cycle_stats.reported_charge_capacity_mah /
                        transformer.cycle_stats.calculated_charge_capacity_mah - 1).iloc[1:-1])).all()


@pytest.mark.transform
@pytest.mark.arbin
@pytest.mark.stats
def test_cccv_data():
    import pandas as pd

    data_path = join(
        CCCV_PATH, 'BG_Arbin_TestData_CCCV_Cell_2_Channel_26_Wb_1.CSV')
    stat_path = join(
        CCCV_PATH, 'BG_Arbin_TestData_CCCV_Cell_2_Channel_26_StatisticByCycle.CSV')
    schedule_path = join(
        CCCV_PATH, 'BG_Arbin_Cell_TestData_CCCV+BG_Arbin_Cell.sdx')

    extractor = Extractor()
    extractor.data_from_files([data_path])
    extractor.data_from_files([stat_path])
    schedule = extractor.schedule_from_files([schedule_path])

    transformer = Transformer()
    transformer.transform_test_data(extractor.raw_test_data)
    transformer.transform_cycle_stats(extractor.raw_cycle_stats)

    transformer.calc_cycle_stats(
        schedule['steps'], cv_voltage_threshold_mv=4200)

    true_values = pd.DataFrame(
        columns=['cycle', 'cc_cap', 'cv_cap', 'cc_time', 'cv_time'])
    true_values['cycle'] = pd.Series([1, 2, 3])
    true_values['cc_cap'] = pd.Series([0, 2567.811, 2561.239])
    true_values['cv_cap'] = pd.Series([0, 6.2E-02, 2.743])
    true_values['cc_time'] = pd.Series([0, 73971.1505, 36904.3838])
    true_values['cv_time'] = pd.Series([0, 1.7623, 60.6568])

    # Make sure calculated CC-CV stats are within 1% of actual value.
    assert ((0.01 > abs(transformer.cycle_stats.calculated_cc_charge_time_s /
                        true_values['cc_time'] - 1).fillna(0))).all()
    assert ((0.01 > abs(transformer.cycle_stats.calculated_cv_charge_time_s /
                        true_values['cv_time'] - 1).fillna(0))).all()
    assert ((0.01 > abs(transformer.cycle_stats.calculated_cc_capacity_mah /
                        true_values['cc_cap'] - 1).fillna(0))).all()
    assert ((0.01 > abs(transformer.cycle_stats.calculated_cv_capacity_mah /
                        true_values['cv_cap'] - 1).fillna(0))).all()

    # Make sure sum of CC-CV values are within 1% of reported total values.
    # Ignore first row because it just has CC step.
    assert ((0.01 > abs(transformer.cycle_stats.reported_charge_time_s.iloc[1:] /
                        (true_values['cc_time'].iloc[1:] + true_values['cv_time'].iloc[1:]) - 1))).all()
    assert ((0.01 > abs(transformer.cycle_stats.reported_charge_capacity_mah.iloc[1:] /
                        (true_values['cc_cap'].iloc[1:] + true_values['cv_cap'].iloc[1:]) - 1))).all()


@pytest.mark.transform
@pytest.mark.arbin
@pytest.mark.stats
def test_capacity_reset_on_step():
    data_path = join(
        RESET_CAP_PATH, 'BG_Maccor_Cell_ResetCap_Take11 - 074.txt')
    stat_path = join(
        RESET_CAP_PATH, 'BG_Maccor_Cell_ResetCap_Take11 - 074 [STATS].txt')
    schedule_path = join(
        RESET_CAP_PATH, 'BG_Maccor_ResetCap.000')

    extractor = Extractor()
    extractor.data_from_files([data_path])
    extractor.data_from_files([stat_path])
    schedule = extractor.schedule_from_files([schedule_path])

    transformer = Transformer()
    transformer.transform_test_data(extractor.raw_test_data)
    transformer.transform_cycle_stats(extractor.raw_cycle_stats)

    transformer.calc_cycle_stats(
        schedule['steps'], cv_voltage_threshold_mv=4429.5)

    test_cycle = 7
    cc_time = 1701.14
    cc_cap = 2120.64
    cv_time = 4541.32
    cv_cap = 1354.83
    discharge_time = 3447.77
    discharge_cap = 3474.54
    charge_time_50 = 1352.0
    charge_time_80 = 2555.99

    # Make sure calculated CC-CV stats are within 1% of actual value.
    assert (0.01 > abs(
        transformer.cycle_stats.at[test_cycle, 'calculated_cc_charge_time_s'] / cc_time - 1))
    assert (0.01 > abs(
        transformer.cycle_stats.at[test_cycle, 'calculated_cv_charge_time_s'] / cv_time - 1))
    assert (0.01 > abs(
        transformer.cycle_stats.at[test_cycle, 'calculated_cc_capacity_mah'] / cc_cap - 1))
    assert (0.01 > abs(
        transformer.cycle_stats.at[test_cycle, 'calculated_cv_capacity_mah'] / cv_cap - 1))
    assert (0.01 > abs(
        transformer.cycle_stats.at[test_cycle, 'calculated_discharge_time_s'] / discharge_time - 1))
    assert (0.01 > abs(
        transformer.cycle_stats.at[test_cycle, 'calculated_discharge_capacity_mah'] / discharge_cap - 1))

    # Make sure 50% and 80% charge times are within 1% of actual value.
    assert (0.1 > abs(
        transformer.cycle_stats.at[test_cycle, 'calculated_fifty_percent_charge_time_s'] / charge_time_50 - 1))
    assert (0.1 > abs(
        transformer.cycle_stats.at[test_cycle, 'calculated_eighty_percent_charge_time_s'] / charge_time_80 - 1))

    # Make sure reported/calculated stats are within 1% of each other.
    # Only look at discharge. Reported charge cap is off because it's reset at each step. Shows the cap from the last charge step.
    assert ((0.01 > abs(transformer.cycle_stats.reported_discharge_capacity_mah /
                        transformer.cycle_stats.calculated_discharge_capacity_mah - 1).iloc[1:-1])).all()


@pytest.mark.transform
@pytest.mark.arbin
@pytest.mark.stats
def test_max_temperature():
    import pandas as pd

    data_path = join(
        ARBIN_SINGLE_PATH, 'BG_Arbin_TestData_Single_File_Channel_26_Wb_1.pkl')
    stat_path = join(
        ARBIN_SINGLE_PATH, 'BG_Arbin_TestData_Single_File_Channel_26_StatisticByCycle.pkl')
    schedule_path = join(
        ARBIN_SINGLE_PATH, 'BG_25R_Characterization+BG_25R.sdx')

    extractor = Extractor()
    df_raw_test_data = extractor.from_pickle(data_path)
    df_raw_cycle_stats = extractor.from_pickle(stat_path)
    schedule = extractor.schedule_from_files([schedule_path])

    transformer = Transformer()
    transformer.transform_test_data(df_raw_test_data)
    transformer.transform_cycle_stats(df_raw_cycle_stats)

    transformer.calc_cycle_stats(
        schedule['steps'], cv_voltage_threshold_mv=4195, cell_thermocouple=2)

    max_temps = pd.DataFrame(
        data=[[1, None, 25.09362588],
              [2, 24.80076, 24.66491629],
              [3, 24.94959932, 25.34211339],
              [4, 25.09418, 26.08771662],
              [5, 26.97403, 28.22165],
              [6, 30.26020972, 31.67171],
              [7, 38.144, 39.89985],
              [8, 47.34957491, 48.98906797],
              [9, 30.24211, None]],
        columns=['cycle', 'max_charge_temp', 'max_discharge_temp'])

    # check that max temperatures are within 0.001 C of expected values
    assert (0.001 > abs(
        transformer.cycle_stats.loc[0:7, 'calculated_max_discharge_temp_c'] - max_temps.loc[0:7, 'max_discharge_temp'])).all()
    assert (0.001 > abs(
        transformer.cycle_stats.loc[1:8, 'calculated_max_charge_temp_c'] - max_temps.loc[1:8, 'max_charge_temp'])).all()


@pytest.mark.transform
@pytest.mark.arbin
@pytest.mark.stats
def test_step_order_stat_calcs():
    data_path = join(
        STEP_ORDER_PATH, 'BG_Arbin_MBC5v2_Cell_Cell6_Channel_25_Wb_1.csv')
    stat_path = join(
        STEP_ORDER_PATH, 'BG_Arbin_MBC5v2_25R_Cell6_Channel_25_StatisticByCycle.CSV')
    schedule_path = join(
        STEP_ORDER_PATH, 'BG_Arbin_MBC5_25R_3+BG_Arbin_25R.sdx')

    extractor = Extractor()
    extractor.data_from_files([data_path])
    extractor.data_from_files([stat_path])
    schedule = extractor.schedule_from_files([schedule_path])

    transformer = Transformer()
    transformer.transform_test_data(extractor.raw_test_data)
    transformer.transform_cycle_stats(extractor.raw_cycle_stats)

    transformer.calc_cycle_stats(
        schedule['steps'], cv_voltage_threshold_mv=4420)

    fifty_percent_charge_time, eighty_percent_charge_time = transformer.cycle_stats.loc[transformer.cycle_stats.cycle == 8, [
        'calculated_fifty_percent_charge_time_s', 'calculated_eighty_percent_charge_time_s']].values[0]

    assert (0.01 > abs(fifty_percent_charge_time / 614.07 - 1))
    assert (0.01 > abs(eighty_percent_charge_time / 1040.08 - 1))



import os
import re
import sys
import json
import subprocess
import pandas as pd

from battetl import logger
from battetl.extract import Extractor
from battetl.transform import Transformer
from battetl.load import QuickLoader


def battetl_quick(file_path: str, file_meta = None) -> pd.DataFrame:
    '''
    The BattETL Quick Mode. Extracts test data or cycle stats from a file, transforms it,
    and loads it to the database.

    Parameters
    ----------
    file_path : str
        Name of the file to be processed.
    file_meta : dict
        A dictionary used to decode unstructured data.

    Returns
    -------
    pd.DataFrame
        The test data or cycle stats that was loaded to the database.
    '''

    extractor = Extractor()
    try:
        extractor.data_from_files([file_path], file_meta)
        raw_test_data = extractor.raw_test_data
        raw_cycle_stats = extractor.raw_cycle_stats
    except Exception as e:
        logger.error('Failed to extract test data', exc_info=True)

    transformer = Transformer()
    test_data = pd.DataFrame()
    cycle_stats = pd.DataFrame()
    if not raw_test_data.empty:
        try:
            transformer.transform_test_data(raw_test_data, file_meta)
            test_data = transformer.test_data
        except Exception as e:
            logger.error('Failed to transform test data', exc_info=True)
    elif not raw_cycle_stats.empty:
        try:
            transformer.transform_cycle_stats(raw_cycle_stats)
            cycle_stats = transformer.cycle_stats
        except Exception as e:
            logger.error('Failed to transform cycle stats', exc_info=True)

    loader = QuickLoader(file_path)
    if not test_data.empty:
        try:
            loader.load_test_data(test_data)
            logger.info(
                f'**** Data successfully loaded to test_data table under test_name {file_path} ****')
        except Exception as e:
            logger.error('Failed to load test data', exc_info=True)
    elif not cycle_stats.empty:
        try:
            loader.load_cycle_stats(cycle_stats)
            logger.info(
                f'**** Data successfully loaded to test_data_cycle_stats table under test_name {file_path} ****')
        except Exception as e:
            logger.error('Failed to load cycle stats', exc_info=True)

    if not test_data.empty:
        return test_data
    elif not cycle_stats.empty:
        return cycle_stats
    else:
        return pd.DataFrame()


def clone_battdb(battdb_folder: str = 'battdb'):
    '''
    Clones the BattDB repo to the specified folder.
    
    Parameters
    ----------
    battdb_folder : str, optional
        Name of the folder to clone battdb to. The default is 'battdb'.
    '''
    if not os.path.exists(battdb_folder):
        logger.info(f'Cloning battdb to {battdb_folder}...')
        subprocess.check_output(
            ['git', 'clone', '--depth', '1', 'https://github.com/BattGenie/battdb'])


def is_docker_installed() -> bool:
    '''
    Checks if Docker is installed.

    Returns
    -------
    bool
    '''
    try:
        logger.info('Checking if Docker is installed...')
        subprocess.check_output(['docker', '-v'])
        return True
    except:
        return False


def is_docker_compose_installed() -> bool:
    '''
    Checks if Docker Compose is installed.

    Returns
    -------
    bool
    '''
    try:
        subprocess.check_output(['docker-compose', '-v'])
        logger.info('Checking if Docker Compose is installed...')
        return True
    except:
        pass

    try:
        subprocess.check_output(['docker', 'compose', '-v'])
        return True
    except:
        pass

    return False


def run_docker_compose(battdb_folder: str = 'battdb') -> subprocess.CompletedProcess:
    '''
    Runs docker-compose up -d with the docker-compose.yml file in the BattDB repo.

    Parameters
    ----------
    battdb_folder : str, optional
        Name of the folder to clone BattDB to. The default is 'battdb'.

    Returns
    -------
    subprocess.CompletedProcess
    '''
    compose_file = os.path.join(
        os.getcwd(), battdb_folder, 'assets', 'battdb_docker', 'docker-compose.yml')
    logger.info('Bringing the database up and applying migrations.')
    logger.info(
        f'Running docker-compose up -d with compose file {compose_file}...')
    env = os.environ.copy()
    result = None
    try:
        result = subprocess.run([
            'docker-compose',
            '-f', compose_file,
            'up', '-d',
        ],
            env=env,
            capture_output=True,
            text=True,
        )
    except:
        result = subprocess.run([
            'docker', 'compose',
            '-f', compose_file,
            'up', '-d',
        ],
            env=env,
            capture_output=True,
            text=True,
        )

    return result


if __name__ == '__main__':
    n = len(sys.argv)
    print("num arguments" + str(n))
    if n < 3:
        print("Not enough arguments passed! Exiting!")
        sys.exit()
    elif n > 4:
        print("Too many arguments passed! Exiting!")
        sys.exit()
    else:
        for i in range(1, n):
            arg = str(sys.argv[i])
            if arg.startswith('file='):
                file_path = re.split('file=', arg)[-1]
            elif arg.startswith('db_url='):
                db_url = re.split('db_url=', arg)[-1]
            elif arg.startswith('file_meta='):
                file_meta_path = re.split('file_meta=', arg)[-1]
            else:
                print("Unknown argument " + arg + "! Exiting!")
                sys.exit()

    if not os.path.isfile(file_path):
        print("File path " + file_path + " does not exist! Exiting!")
        sys.exit()

    file_meta = None
    if os.path.isfile(file_meta_path):
        with open(file_meta_path) as file:
            file_meta = json.load(file)
    else:
        print("Invalid path for file_meta! Exiting!")
        sys.exit()

    # Assumes a db_url format of:
    # "postgres://YourUserName:YourPassword@YourHostname:Port/YourDatabaseName"
    split_list = re.split('//', db_url)[-1].split(':')
    username = split_list[0]
    password, hostname = split_list[1].split("@")
    port, dbname = split_list[2].split("/")
    battdb_folder = 'battdb'

    # Set the env variables based on the passed info.
    os.environ['DB_USERNAME'] = username
    os.environ['DB_PASSWORD'] = password
    os.environ['DB_HOSTNAME'] = hostname
    os.environ['DB_PORT'] = port
    os.environ['DB_TARGET'] = dbname
    # For docker-compose
    os.environ['POSTGRES_USER'] = username
    os.environ['POSTGRES_PASSWORD'] = password
    os.environ['POSTGRES_PORT'] = port
    os.environ['POSTGRES_DATABASE'] = dbname
    flyway_sql = os.path.join(
        os.getcwd(), battdb_folder, 'assets', 'migration_scripts_quick')
    os.environ['FLYWAY_SQL'] = flyway_sql

    if not is_docker_installed():
        print(f'Docker is not installed. Please go to https://www.docker.com/ to install Docker.')
        sys.exit()

    if not is_docker_compose_installed():
        print(f'Docker Compose is not installed. Please go to https://docs.docker.com/compose/ to install Docker Compose.')
        sys.exit()

    clone_battdb(battdb_folder)
    docker_result = run_docker_compose(battdb_folder)
    logger.info(docker_result.stdout)
    logger.info(docker_result.stderr)

    battetl_quick(file_path, file_meta)
